import cv2

def padding_image(image, x1, y1, x2, y2, padding=5):
    h, w = image.shape[:2]
    x1 = max(x1 - padding, 0)
    y1 = max(y1 - padding, 0)
    x2 = min(x2 + padding, w)
    y2 = min(y2 + padding, h)
    return x1, y1, x2, y2



# my_face_library/detection.py

import cv2
from ultralytics import YOLO
import numpy as np

from .config import DEVICE, DEFAULT_YOLO_MODEL_PATH, DEFAULT_CONFIDENCE_THRESHOLD, DEFAULT_IOU_THRESHOLD, DEFAULT_PADDING
from .utils import padding_image

_yolo_model = None

def _get_yolo_model(model_path: str = DEFAULT_YOLO_MODEL_PATH):
    global _yolo_model
    if _yolo_model is None:
        _yolo_model = YOLO(model_path).to(DEVICE)
    return _yolo_model


def face_detection(
    image_bgr: np.ndarray,
    conf: float = DEFAULT_CONFIDENCE_THRESHOLD,
    iou: float = DEFAULT_IOU_THRESHOLD,
    do_track: bool = False,
    tracker_config: str = "bytetrack.yaml",
):

    model = _get_yolo_model()
    
    if do_track:
        results = model.track(
            image_bgr,
            conf=conf,
            iou=iou,
            persist=True,
            tracker=tracker_config,
            verbose=False
        )
    else:
        results = model.predict(
            image_bgr,
            conf=conf,
            iou=iou,
            verbose=False
        )
    
    detections = []
    for result in results:
        for box in result.boxes:
            x1, y1, x2, y2 = box.xyxy[0]
            x1, y1, x2, y2 = padding_image(image_bgr, x1, y1, x2, y2, padding=DEFAULT_PADDING)
            
            score = float(box.conf[0]) if box.conf is not None else None
            track_id = int(box.id[0].item()) if box.id is not None else None
            detections.append({
                "box": (int(x1), int(y1), int(x2), int(y2)),
                "track_id": track_id,
                "score": score
            })
    
    return detections




import torch
import os

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"

DEFAULT_YOLO_MODEL_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "model")
YOLO_CHECKPOINT = os.path.join(DEFAULT_YOLO_MODEL_PATH, "yolov11m-face.pt")


DEFAULT_DB_PATH = "./"
DEFAULT_COLLECTION_NAME = "face_collection"
DEFAULT_DIMENSION = 512
DEFAULT_METRIC = "COSINE"
DEFAULT_CONFIDENCE_THRESHOLD = 0.4
DEFAULT_IOU_THRESHOLD = 0.4

DEFAULT_PADDING = 5



from pathlib import Path
from typing import Optional, Tuple

import cv2
import numpy as np
import numpy.typing as npt
import onnxruntime as ort

from .feature_extractor import FeatureExtractor

DEFAULT_MODELS_PATH = Path(__file__).parent / "models"


class EdgeFaceFeatureExtractor(FeatureExtractor):
    def __init__(
        self,
        input_shape: Optional[Tuple[int, int]] = (112, 112),
        model_path: Optional[Path] = DEFAULT_MODELS_PATH / "edgeface_s_gamma_05.onnx",
    ) -> None:
        """Initializes a EdgeFaceFeatureExtractor object from a provided path to an ONNX file. This
        class also includes all the preprocessing steps required by the EdgeFace model.

        Model reference: https://gitlab.idiap.ch/bob/bob.paper.tbiom2023_edgeface

        Args:
            input_shape (Optional[Tuple[int, int]]): Shape of the input image. Defaults to (112, 112).
            model_path (Optional[Path]): Path to the ONNX model file. Defaults to the EdgeFace model.
        """
        self.input_shape = input_shape
        self._init_model(model_path=model_path)

    def _init_model(self, model_path: str) -> None:
        """Initializes the ONNX model by loading it from the provided model path. The model is
        initialized as an onnxruntime InferenceSession object.

        Args:
            model_path (str): Path to the ONNX model file.
        """
        self.ort_session = ort.InferenceSession(model_path)

    def _preprocess(self, image: npt.NDArray) -> npt.NDArray:
        """Performs all preprocessing steps before predicting over an image. The EdgeFace model
        requires an input of size 112x112, with torch style norm,

        (img / 255. - 0.5) / 0.5, and with order BxCxWxH

        Args:
            image (npt.NDArray): Image to perform preprocessing on, shape (H, W, C) is expected.

        Returns:
            npt.NDArray: An object representing the image as required by the actual model, with shape
            (1, C, H, W).
        """
        # resize image
        # preprocessed_image = cv2.resize(image, self.input_shape).astype(np.float32)
        preprocessed_image = cv2.resize(image, self.input_shape, interpolation=cv2.INTER_LANCZOS4).astype(np.float32)
        # 0 - 255 to 0.0 - 1.0 with mean and std of 0.5
        preprocessed_image = (preprocessed_image / 255.0 - 0.5) / 0.5
        # HWC to CHW
        preprocessed_image = preprocessed_image.transpose((2, 0, 1))
        # adding batch dim
        preprocessed_image = np.expand_dims(preprocessed_image, axis=0)

        return preprocessed_image

    def predict(self, image: npt.NDArray) -> npt.NDArray:
        """Predicts the embeddings for the given image. Note that the image is expected to be a
        crop of a face, since this model is not responsible for face detection.

        Args:
            image (npt.NDArray): The input image, with shape (H, W, C).

        Returns:
            npt.NDArray: The embeddings of the image, with shape (512,).
        """
        preprocessed_image = self._preprocess(image)
        embeddings = self.ort_session.run(None, {"data": preprocessed_image})[0]

        # removing batch dim
        embeddings = embeddings.squeeze()

        return embeddings



from pymilvus import MilvusClient
import numpy as np
import time
import os

from .config import DEFAULT_DB_PATH, DEFAULT_COLLECTION_NAME, DEFAULT_DIMENSION, DEFAULT_METRIC

def db_set(
    db_path: str = DEFAULT_DB_PATH,
    collection_name: str = DEFAULT_COLLECTION_NAME,
    dimension: int = DEFAULT_DIMENSION,
    metric_type: str = DEFAULT_METRIC,
    auto_id: bool = True,
    enable_dynamic_field: bool = True
):

    client = MilvusClient(uri=db_path)
    if not client.has_collection(collection_name=collection_name):
        client.create_collection(
            collection_name=collection_name,
            vector_field_name="vector",
            dimension=dimension,
            auto_id=auto_id,
            enable_dynamic_field=enable_dynamic_field,
            metric_type=metric_type,
        )
    return client


def db_insert(
    client: MilvusClient,
    collection_name: str,
    vector: np.ndarray,
    orig_path: str,
    crop_path: str,
    timestamp: str,
    tracking_id: int,
    cam_id: str
):

    if not isinstance(vector, list):
        vector = vector.tolist()

    data_dict = {
        "vector": vector,
        "orig_path": orig_path,
        "crop_path": crop_path,
        "timestamp": timestamp,
        "tracking_id": tracking_id,
        "cam_id": cam_id,
    }
    client.insert(collection_name=collection_name, data=data_dict)
    return True

def db_search(
    client: MilvusClient,
    collection_name: str,
    query_image: np.ndarray,
    top_k: int = 5,
    threshold: float = 0.4,
    extractor_func=None,
):

    if extractor_func is None:
        raise ValueError("extractor_func is required. (e.g. face_extraction)")
    query_vector = extractor_func(query_image)
    if not isinstance(query_vector, list):
        query_vector = query_vector.tolist()

    results = client.search(
        collection_name=collection_name,
        data=[query_vector],
        output_fields=[
            "timestamp", 
            "orig_path", 
            "crop_path", 
            "tracking_id", 
            "cam_id"
        ],
        param={"metric_type": "COSINE"},
        limit=top_k
    )

    hits = []
    for hit in results[0]:
        score = hit.distance
        if score >= threshold:
            hits.append({
                "score": score,
                "entity": hit.entity
            })

    return hits
